from typing import Optional, List

from pydantic import BaseModel, ConfigDict, Field
from kodexa.model.base import StandardDateTime
from kodexa.model.objects import User
from kodexa.platform.client import EntityEndpoint, PageEndpoint, EntitiesEndpoint

from enum import Enum


class CheckStatus(Enum):
    """ Check Status ENUM: OPEN, CLOSED, IN_PROGRESS """
    open = "OPEN"
    closed = "CLOSED"
    in_progress = "IN_PROGRESS"


class CheckResponse(BaseModel):
    """
    Entity of check response
    """
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        protected_namespaces=("model_config",),
    )

    id: Optional[str] = Field(None)
    uuid: Optional[str] = None
    change_sequence: Optional[int] = Field(None, alias="changeSequence")
    created_on: Optional[StandardDateTime] = Field(None, alias="createdOn")
    updated_on: Optional[StandardDateTime] = Field(None, alias="updatedOn")
    title: Optional[str] = None
    description: Optional[str] = None
    status: Optional[CheckStatus] = None
    confidence: Optional[float] = None
    approver: Optional[User] = None


class PageCheckResponse(BaseModel):
    """
        A page pydantic for check response
    """

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        protected_namespaces=("model_config",),
    )
    total_pages: Optional[int] = Field(None, alias="totalPages")
    total_elements: Optional[int] = Field(None, alias="totalElements")
    size: Optional[int] = None
    content: Optional[List[CheckResponse]] = None
    number: Optional[int] = None
    number_of_elements: Optional[int] = Field(None, alias="numberOfElements")
    first: Optional[bool] = None
    last: Optional[bool] = None
    empty: Optional[bool] = None


class PageCheckResponseEndpoint(PageCheckResponse, PageEndpoint):
    """
        Represents a page check response endpoint

        This class is used to represent the endpoints of a page check response. It inherits from
        the PageCheckResponse and PageEndpoint classes.

        Methods:
            get_type: Returns the type of the endpoint.
    """

    def get_type(self) -> str:
        """Get the type of the endpoint

        This method is used to get the type of the endpoint. In this case, it will always
        return "workspace".

        Returns:
            str: The type of the endpoint, "workspace".
        """
        return "checkResponse"


class CheckResponseEndpoint(CheckResponse, EntityEndpoint):
    """Represents a Check Response endpoint"""

    def get_type(self) -> str:
        """
        Get the type of endpoint

        :return: The type of endpoint
        """
        return "checkResponses"


class CheckResponsesEndpoint(EntitiesEndpoint):
    """ Represents check responses endpoint """

    def get_type(self) -> str:
        """
            Get the type of endpoint
        :return: The type of endpoint
        """
        return "checkResponses"

    def get_instance_class(self, object_dict=None) -> CheckResponseEndpoint:
        """Get the instance class of the endpoint

        This method is used to get the instance class of the endpoint.

        Args:
            object_dict (dict, optional): A dictionary containing the object data.

        Returns:
            CheckResponseEndpoint: The instance class of the endpoint.
        """
        return CheckResponseEndpoint

    def get_page_class(self, object_dict=None) -> PageCheckResponseEndpoint:
        """Get the page class of the endpoint

        This method is used to get the page class of the endpoint.

        Args:
            object_dict (dict, optional): A dictionary containing the object data.

        Returns:
            PageCheckResponseEndpoint: The page class of the endpoint.
        """
        return PageCheckResponseEndpoint



import logging

from kodexa import ContentNode

logger = logging.getLogger(__name__)

def get_pretty_text_from_lines(lines: list[ContentNode], scale, include_line_uuid=False) -> str:
    pretty_text = ""
    for line_index, line in enumerate(lines):
        line_content = f"('{line.uuid}')" if include_line_uuid else ""
        current_x = 0
        for word in line.select('//word'):
            x = int(word.get_bbox()[0] * scale)
            spaces_needed = max(1, x - current_x)  # Ensure at least one space
            line_content += " " * spaces_needed
            line_content += f"{word.get_all_content()}"
            current_x = x + len(word.get_all_content())

        pretty_text += line_content + "\n"

    return pretty_text


def get_max_width(lines: list[ContentNode], max_width=None) -> int:
    if max_width is None:
        if lines is None or len(lines) == 0:
            return 250
        
        # Find the line with the most words
        max_words_line = max(lines, key=lambda line: sum(len(word.get_all_content()) for word in line.select('//word')))

        # Calculate max_width based on the length of all words plus spaces
        max_width = sum(len(word.get_all_content()) for word in max_words_line.select('//word')) + (len(max_words_line.select('//word'))*4) - 1

    if max_width < 250:
        max_width = 250

    return max_width


def get_scale_from_words(words: list[ContentNode], max_width) -> float:
    # Get the bboxes
    bboxes = [word.get_bbox() for word in words]

    # Find the overall bounding box
    min_x = min(bbox[0] for bbox in bboxes)
    max_x = max(bbox[2] for bbox in bboxes)
    min_y = min(bbox[1] for bbox in bboxes)
    max_y = max(bbox[3] for bbox in bboxes)

    # Invert y-axis
    max_y, min_y = min_y, max_y

    # Calculate scale factor to fit within max_width
    scale = max_width / (max_x - min_x)

    return scale


def get_pretty_page(page: ContentNode, max_width=None, include_line_uuid=False) -> str:
    """
    Get a pretty representation of the page

    :param page: The page to get the pretty representation for
    :param max_width: The maximum width of the page
    :param include_line_uuid: Include the line UUID in the pretty representation

    :return: A pretty representation of the page
    """

    logger.info(f"Getting pretty page {page.index}")

    pretty_text = ""
    content_areas = page.select('//content-area')

    lines = page.select('//line')

    max_width = get_max_width(lines, max_width)
    logger.info(f"Max width: {max_width}")

    words = page.select('//word')
    if len(words) == 0:
        return page.get_all_content()

    scale = get_scale_from_words(words, max_width)
    for area_index, area in enumerate(content_areas):

        if area_index > 0:
            pretty_text += "\n\n"  # Add extra newline between content areas

        pretty_text += get_pretty_text_from_lines(area.select('//line'), scale, include_line_uuid)

    logger.debug(f"Pretty Page: {page.index}: \n{pretty_text}")

    return pretty_text


from decimal import Decimal
from typing import Optional, List, Set

from pydantic import BaseModel, ConfigDict, Field

from kodexa.model.base import StandardDateTime
from kodexa.platform.client import EntityEndpoint, PageEndpoint, EntitiesEndpoint
from .product_group import ProductGroup
from ..objects import ProjectTemplate


class ProjectTemplateMetadata(BaseModel):
    """
    A project template metadata entity
    """
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        protected_namespaces=("model_config",),
    )

    id: str


class ProductProjectTemplate(BaseModel):
    """
    A product project template entity representing the relationship between products and project templates
    """
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        protected_namespaces=("model_config",),
    )

    id: Optional[str] = None
    uuid: Optional[str] = None
    change_sequence: Optional[int] = Field(None, alias="changeSequence")
    created_on: Optional[StandardDateTime] = Field(None, alias="createdOn")
    updated_on: Optional[StandardDateTime] = Field(None, alias="updatedOn")
    display_order: Optional[int] = Field(None, alias="displayOrder")
    project_template_metadata: Optional[ProjectTemplateMetadata] = Field(None, alias="projectTemplateMetadata")


class Product(BaseModel):
    """
    A product entity representing a product in the Kodexa platform
    """
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        protected_namespaces=("model_config",),
    )

    id: Optional[str] = None
    uuid: Optional[str] = None
    change_sequence: Optional[int] = Field(None, alias="changeSequence")
    created_on: Optional[StandardDateTime] = Field(None, alias="createdOn")
    updated_on: Optional[StandardDateTime] = Field(None, alias="updatedOn")
    name: str
    description: Optional[str] = None
    overview_markdown: Optional[str] = Field(None, alias="overviewMarkdown")
    product_group: ProductGroup = Field(..., alias="productGroup")
    parent: Optional['Product'] = None
    image_url: Optional[str] = Field(None, alias="imageUrl")
    price_id: Optional[str] = Field(None, alias="priceId")
    price: Optional[Decimal] = None
    number_of_credits: Optional[int] = Field(None, alias="numberOfCredits")
    price_suffix: Optional[str] = Field(None, alias="priceSuffix")
    has_quantity: bool = Field(False, alias="hasQuantity")
    active: bool = True
    order: Optional[int] = None
    promoted: Optional[bool] = None
    project_templates: Optional[Set[ProjectTemplate]] = Field(None, alias="projectTemplates")
    search_text: Optional[str] = None

    def update_search_text(self):
        """Updates the search text for the product"""
        if self.product_group:
            self.search_text = f"{self.name.lower()} {self.product_group.name.lower()}"
        else:
            self.search_text = self.name.lower()


class ProductEndpoint(Product, EntityEndpoint):
    """Handles the endpoint for a product

    This class is a combination of DataException and EntityEndpoint. It is used
    to manage the endpoint for data exceptions.

    Methods:
        get_type: Returns the type of the endpoint.
    """

    def get_type(self) -> str:
        """Gets the type of the endpoint.

        This method returns the type of the endpoint which is "exceptions".

        Returns:
            str: The type of the endpoint.
        """
        return "products"


class PageProduct(BaseModel):
    """
    Represents a paginated list of products
    """
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        protected_namespaces=("model_config",),
    )
    total_pages: Optional[int] = Field(None, alias="totalPages")
    total_elements: Optional[int] = Field(None, alias="totalElements")
    size: Optional[int] = None
    content: Optional[List[Product]] = None
    number: Optional[int] = None
    number_of_elements: Optional[int] = Field(None, alias="numberOfElements")
    first: Optional[bool] = None
    last: Optional[bool] = None
    empty: Optional[bool] = None


class PageProductEndpoint(PageProduct, PageEndpoint):
    def get_type(self) -> Optional[str]:
        return "product"


class ProductsEndpoint(EntitiesEndpoint):
    """Represents the products endpoint

    This class is used to represent the products endpoint in the system.

    Attributes:
        object_dict: A dictionary containing the object data.
    """

    """Represents a assistants endpoint"""

    def get_type(self) -> str:
        """Get the type of the endpoint

        This method is used to get the type of the endpoint.

        Returns:
            str: The type of the endpoint.
        """
        return "products"

    def get_instance_class(self, object_dict=None):
        """Get the instance class of the endpoint

        This method is used to get the instance class of the endpoint.

        Args:
            object_dict (dict, optional): A dictionary containing the object data.

        Returns:
            AssistantEndpoint: The instance class of the endpoint.
        """
        return ProductEndpoint

    def get_page_class(self, object_dict=None):
        """Get the page class of the endpoint

        This method is used to get the page class of the endpoint.

        Args:
            object_dict (dict, optional): A dictionary containing the object data.

        Returns:
            PageAssistantEndpoint: The page class of the endpoint.
        """
        return PageProductEndpoint
